/*******************************************************************************
   RoBoard ADC Simple Tool

     This program is for the purpose of demonstrating RoBoard ADC function. 
     Execute it without arguments to see how to use.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if defined(linux)
    #define stricmp  strcasecmp
#endif

#ifdef WINCE
    #include <windows.h>
	#define stricmp _stricmp
#endif

#if defined(WINCE) || defined(WIN32)
    #include <roboard_dll.h>  // use the DLL version of RoBoIO lib
    // #include <roboard.h>   // use the static version of RoBoIO lib
#else
    #include <roboard.h>
#endif


void help(char* s) {
    int l = strlen(s);

    for (; l>=0; l--) if ((s[l] == '/') || (s[l] == '\\')) break;
    if (l >= 0) s = &s[l+1];

	printf(
	    "Usage                                         \n"
	    "=======                                       \n"
	    "%s [ver] [range] [coding]                     \n"
	    "                                              \n"
	    "   [ver] (RoBoard version)                    \n"
	    "      RB050   : RoBoard RB-050                \n"
	    "      RB100   : RoBoard RB-100                \n"
	    "      RB100RD : RoBoard RB-100RD              \n"
	    "      RB110   : RoBoard RB-110                \n"
	    "                                              \n"
	    "   [range] (input voltage range):             \n"
	    "      0 : 0 ~ 2.5V                            \n"
	    "      1 : 0 ~ 5V                              \n"
	    "                                              \n"
	    "   [coding] (ADC coding range):               \n"
	    "      0 : Unsigned Coding                     \n"
	    "          (ex. AD7918: 0 ~ 1023)              \n"
	    "      1 : Signed Coding                       \n"
	    "          (ex. AD7918: -511 ~ 512)            \n"
	    "                                              \n"
	    "(Press Enter to continue...)                  \n"
	    , s
    ); 
    getchar();

	printf(
	    "example                                                        \n"
	    "=========                                                      \n"
	    "%s RB100 0 0                                                   \n"
	    "    read ADC values on RB-100 with voltage range 0V ~ 2.5V and \n"
	    "    coding range 0 (for 0V) ~ 1023 (for 2.5V).                 \n"
	    "                                                               \n"
	    "%s RB110 1 0                                                   \n"
	    "    read ADC values on RB-110 with voltage range 0V ~ 5V and   \n"
	    "    coding range 0 (for 0V) ~ 1023 (for 5V).                   \n"
	    "                                                               \n"
	    , s, s
    );
}

int GetADCRange(char* str) {
	if(strcmp("0", str) == 0)
		return ADCMODE_RANGE_VREF;
	else
	if(strcmp("1", str) == 0)
		return ADCMODE_RANGE_2VREF;

    printf("ERR: wrong voltage range!!!\n\n");
    exit(1);
}


int GetADCCoding(char* str) {
	if(strcmp("0", str) == 0)
		return ADCMODE_UNSIGNEDCODING;
	else
	if(strcmp("1", str) == 0)
		return ADCMODE_SIGNEDCODING;

	printf("ERR: wrong coding range!!!\n\n");
    exit(1);
}

void SetRBVer(int argc, char* argv[]) {
    if (stricmp(argv[1], "RB050") == 0)
    {
        roboio_SetRBVer(RB_050); return;
    }
    else
    if (stricmp(argv[1], "RB100") == 0)
    {
        roboio_SetRBVer(RB_100); return;
    }
    else
    if (stricmp(argv[1], "RB100b1") == 0)
    {
        roboio_SetRBVer(RB_100b1); return;
    }
    else
    if (stricmp(argv[1], "RB100b2") == 0)
    {
        roboio_SetRBVer(RB_100b2); return;
    }
    else
    if (stricmp(argv[1], "RB100RD") == 0)
    {
        roboio_SetRBVer(RB_100RD); return;
    }
    else
    if (stricmp(argv[1], "RB110") == 0)
    {
        roboio_SetRBVer(RB_110); return;
    }

	printf("ERR: unknown RoBoard version!!!\n\n");
    exit(1);
}

#ifdef WINCE
int _tmain(int argc, _TCHAR* _argv[])
{
	char  argvbuf[50][100] = {'\0'};
	char* argv[50];

	if (argc > 50) argc = 50;
	for (int _argc=0; _argc<argc; _argc++)
	{
		argv[_argc] = &(argvbuf[_argc][0]);
		wcstombs(argv[_argc], _argv[_argc], 100);
	}
#else
int main(int argc, char* argv[])
{
#endif

	int i, range, coding, val;
	double vol;

	if(argc == 1)
	{
		help(argv[0]);
		return 0;
	}

	if(argc != 4)
	{
		printf("ERR: wrong number of parameters!!!\n\n");
		return 1;
	}

	SetRBVer(argc, argv);

	range  = GetADCRange(argv[2]);
	coding = GetADCCoding(argv[3]);


	// Before using AD7918 functions, you should first initialize SPI interface.
    if (spi_Init(SPICLK_21400KHZ) == false)
	{
        printf("ERR: fail to initialize SPI (%s)!\n", roboio_GetErrMsg());
		return 1;
	}

	for(i=0; i<8; i++)
	{
		val = adc_ReadChannel(i, range, coding);  // read A/D channel i
		if (val == ADC_READFAIL)
            printf("ERR: fail to read ADC channel %d\n", i);
		else
		{
		    vol = (coding == ADCMODE_SIGNEDCODING)? (double)(val + 512) : (double)val;
		    vol = vol * ((range == ADCMODE_RANGE_VREF)? 2.5 : 5.0) / 1023.0;
            printf("ADC value on channel %d = %5d (%2.3fV)\n", i, val, vol);
        }
	}

	// Close SPI interface
	spi_Close();
	return 0;
}

