/*******************************************************************************
   RoBoard SPI Simple Tool

     This program is for the purpose of demonstrating RoBoard's native SPI function. 
     Execute it without arguments to see how to use.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if defined(linux)
    #define stricmp  strcasecmp
#endif

#ifdef WINCE
    #include <windows.h>
	#define stricmp _stricmp
#endif

#if defined(WINCE) || defined(WIN32)
    #include <roboard_dll.h>  // use the DLL version of RoBoIO lib
    // #include <roboard.h>   // use the static version of RoBoIO lib
#else
    #include <roboard.h>
#endif

#define HARDWARE		 (0)
#define SOFTWARE		 (1)

#define READ             (2)
#define WRITE            (3)
#define EXCHANGE         (4)

#define BUFF_SIZE        (128)
unsigned int buf[BUFF_SIZE];

void help(char* s) {
    int l = strlen(s);

    for (; l>=0; l--) if ((s[l] == '/') || (s[l] == '\\')) break;
    if (l >= 0) s = &s[l+1];

	printf(
	    "Usage                                                                  \n"
	    "=======                                                                \n"
		"%s [ver] [HW|SW] [-r count |-w byte byte... |-x byte byte ...]         \n"
		"                                                                       \n"
	    "   [ver] (RoBoard version)                                             \n"
	    "      RB100   : RoBoard RB-100                                         \n"
	    "      RB100RD : RoBoard RB-100RD                                       \n"
	    "      (note: only RB-100/RB-100RD is supported)                        \n"
		"                                                                       \n"
		"   [HW|SW] (hardware/software SPI mode):                               \n"
		"      HW : hardware SPI mode (half-duplex),                            \n"
		"           select spi clock = 10MHZ                                    \n"
		"      SW : software-simulated SPI mode (full-duplex),                  \n"
		"           select spi clock ~ 167kHZ and clock mode = [CPOL 0, CPHA 0] \n"
		"                                                                       \n"
		"   [-r count]: half-duplex SPI read [count] bytes from the SPI bus.    \n"
		"                                                                       \n"
		"   [-w byte1 byte2 ...]: half-duplex SPI write byte1, byte2, ... to    \n"
		"                         the SPI bus.                                  \n"
		"                                                                       \n"
		"   [-x byte1 byte2 ...]: full-duplex SPI write byte1, byte2, ... to    \n"
		"                         the SPI bus. (only for software SPI mode)     \n"
		"                                                                       \n"
		"   Note that byte1, byte2, ... should be represented in HEX numbers.   \n"
		"                                                                       \n"
		"(Press Enter key to continue...)                                       \n"
		, s
	);
	getchar();

	printf(
		"example                                                                \n"
		"=========                                                              \n"
		"%s RB100 HW -w 55 aa 12 3b                                             \n"
		"    hardware SPI mode, write 0x55 0xaa 0x12 0x3b to the SPI bus.       \n"
		"                                                                       \n"
		"%s RB100 HW -r 6                                                       \n"
		"    hardware SPI mode, read 6 bytes from the SPI bus.                  \n"
		"                                                                       \n"
		"%s RB100 SW -x 55 aa                                                   \n"
		"    software SPI mode, write 0x55 0xaa to the SPI bus,                 \n"
		"    and read 2 bytes at the same time.                                 \n"
		"                                                                       \n"
		, s, s, s
	);
}

unsigned int htoi(char* str) {
    int i;
	unsigned int value = 0;

	for (i=0; i<(int)strlen(str); i++)
	{
		if(str[i]>='0' && str[i]<='9')
		{
			value = (value<<4)+(unsigned int)(str[i]-'0');
		}
		else
		if(str[i]>='a' && str[i]<='f')
		{
			value = (value<<4)+(unsigned int)(str[i]-'a'+10);
		}
		else
		if(str[i]>='A' && str[i]<='F')
		{
			value = (value<<4)+(unsigned int)(str[i]-'A'+10);
		}
		else
		{
			return 0xffff;
		}
	}

	return value;
}


int GetMode(char* str) {
	if(stricmp("HW", str) == 0)
		return HARDWARE;
	else
	if(stricmp("SW", str) == 0)
		return SOFTWARE;
	
	printf("ERR: wrong SPI HW/SW mode parameter!!!\n\n");
	exit(1);
}

int GetFunction(char* str, int mode) {
	if(stricmp("-w", str) == 0)
		return WRITE;
	else
	if(stricmp("-r", str) == 0)
		return READ;
	else
	if(stricmp("-x", str) == 0)
	{
		if(mode == HARDWARE)
		{
			printf("ERR: [-x] option is only for software SPI mode!!!\n\n");
			exit(1);
		}

		return EXCHANGE;
	}

	printf("ERR: unknown r/w mode parameter!!!\n\n");
	exit(1);
}

int GetReadCount(char* str) {
	int val;

	val = atoi(str);

	if(val <= 0)
	{
		printf("ERR: wrong read count!!!\n\n");
		exit(1);
	}

	if(val > BUFF_SIZE)
	{
		printf("ERR: read count is too large!!!\n\n");
		exit(1);
	}

    return val;
}

void SetRBVer(int argc, char* argv[]) {
    if (stricmp(argv[1], "RB100") == 0)
    {
        roboio_SetRBVer(RB_100); return;
    }
    else
    if (stricmp(argv[1], "RB100b1") == 0)
    {
        roboio_SetRBVer(RB_100b1); return;
    }
    else
    if (stricmp(argv[1], "RB100b2") == 0)
    {
        roboio_SetRBVer(RB_100b2); return;
    }
    else
    if (stricmp(argv[1], "RB100RD") == 0)
    {
        roboio_SetRBVer(RB_100RD); return;
    }

	printf("ERR: unsupported RoBoard version!!!\n\n");
    exit(1);
}

#ifdef WINCE
int _tmain(int argc, _TCHAR* _argv[])
{
	char  argvbuf[50][100] = {'\0'};
	char* argv[50];

	if (argc > 50) argc = 50;
	for (int _argc=0; _argc<argc; _argc++)
	{
		argv[_argc] = &(argvbuf[_argc][0]);
		wcstombs(argv[_argc], _argv[_argc], 100);
	}
#else
int main(int argc, char* argv[])
{
#endif

	int i, mode, status, count;

	if(argc == 1)
	{
		help(argv[0]);
		return 0;
	}

	if(argc < 5)
	{
		printf("ERR: too less parameters!!!\n\n");
		return 1;
	}

	SetRBVer(argc, argv);

	mode   = GetMode(argv[2]);
	status = GetFunction(argv[3], mode);

	if((status == WRITE) || (status == EXCHANGE))
	{
        if((argc-4) > BUFF_SIZE)
		{
			printf("ERR: too many number of bytes to write!!!\n\n");
			return 1;
		}

		for(i=0; i<argc-4; i++)
		{
		    buf[i] = htoi(argv[i+4]);
		    
		    if (buf[i] > 255)
		    {
                printf("ERR: there exists a wrong byte to write!!!\n\n");
                return 1;
            }
        }
    }
    else // if (status == READ)
    {
		if(argc > 5)
		{
			printf("ERR: too many parameters!!!\n\n");
			return 1;
		}

		count = GetReadCount(argv[4]);
		if (count > BUFF_SIZE)
		{
			printf("ERR: too many number of bytes to read!!!\n\n");
			return 1;
		}
	}
    

    if (((mode == HARDWARE)? spi_Init(SPICLK_10000KHZ) : spi_InitSW(SPIMODE_CPOL0 + SPIMODE_CPHA0, 0L)) == false)
    {
        printf("ERR: fail to initialize SPI (%s)!\n", roboio_GetErrMsg());
        return 1;
    }

	if(status == WRITE)
	{
        printf("Write ");
		for(i=0; i<argc-4; i++) printf("%02Xh ", buf[i]);
		printf("to SPI bus...");

		spi_EnableSS();
        for(i=0; i<argc-4; i++)
            if (i < argc-5) spi_Write((unsigned char)buf[i]); else spi_WriteFlush((unsigned char)buf[i]);
		spi_DisableSS();

		printf("completed\n");
	}
	else
	if(status == READ)
	{
		spi_EnableSS();
		for(i=0; i<count; i++) buf[i] = spi_Read();
		spi_DisableSS();

		for(i=0; i<count; i++)
		{
            if (buf[i] == 0xffff)
			    printf("Read byte %d = fail!\n", i);
			else
			    printf("Read byte %d = %02Xh\n", i, buf[i]);
		}
	}
	else
	{
		printf("Exchange ");
		for(i=0; i<argc-4; i++) printf("%02Xh ", buf[i]);
		printf("with SPI bus...");

		spi_EnableSS();
        for(i=0; i<argc-4; i++) buf[i] = spi_Exchange((unsigned char)buf[i]);
		spi_DisableSS();

		printf("completed\n");

		for(i=0; i<argc-4; i++)
		{
            if (buf[i] == 0xffff)
			    printf("Read byte %d = fail!\n", i);
			else
			    printf("Read byte %d = %02Xh\n", i, buf[i]);
		}
	}

	if(mode == HARDWARE)
		spi_Close();
	else
		spi_CloseSW();
	
	return 0;
}

